// $Id: FTPlayerProxy.java,v 1.1 2003/03/06 18:19:09 lw Exp $
package team5.casino.comm;

import org.omg.CORBA.SystemException;
import team5.base.comm.*;

/**
 * This class acts as in intermediary between the dealer and the player
 * CORBA stubs. All calls from the PlayerOperations interface are forwarded
 * to the corresponding Player stub.
 * Each player should have access to exactly one floor, therefore this
 * class is implemented as a singleton.
 *
 * @author Team 5
 * @version $Revision: 1.1 $
 */
public class FTPlayerProxy {

    // Public methods

    /**
     * Get the player proxy object. Create it if necessary.
     * @return the player proxy
     */
    public synchronized static FTPlayerProxy getInstance(Player player) {

        if (theInstance == null) {
            theInstance = new FTPlayerProxy(player);
        }
        return theInstance;
    }


    // Player operations

    // TODO: Fix comments for player operations
    /**
     * A dealer request for a player to bet on his/her cards.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: The player is prompted for his bet for the next game.</li>
     * <li>Clients: Dealer</li></ul>
     * @return The client's bet in chips.
     * @throws CommException
     */
    public int placeBet() throws CommException {

        int bet = 0;

        try {
            bet = remotePlayer.placeBet(
                    ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
        return bet;
    }

    /**
     * A dealer sends a card to a player
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: This method passes a card to the player.</li>
     * <li>Clients: Dealer</li></ul>
     * @param cardID The ID of the new card.
     * @throws CommException
     */
    public void acceptCard(byte cardID) throws CommException {

        try {
            remotePlayer.acceptCard(
                    ObjectID.getOID(), SequenceNumber.next(), cardID);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer is giving a player another player's card to view.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to register for a session
     *           at the casino server.</li>
     * <li>Clients: Dealer</li></ul>
     * @param playerNumber The number of the other player at the
     *                table who gets this card.
     * @param cardID The ID of the new card.
     * @throws CommException
     */
    public void acceptOthersCard(int playerNumber, byte cardID)
            throws CommException {

        try {
            remotePlayer.acceptOthersCard(ObjectID.getOID(),
                    SequenceNumber.next(), playerNumber, cardID);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer is telling the player that the game is starting
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: Start a new game at the current table.</li>
     * <li>Clients: Dealer</li></ul>
     * @param playerCount The number of players at the table.
     * @throws CommException
     */
    public void startGame(byte playerCount) throws CommException {

        try {
            remotePlayer.startGame(
                    ObjectID.getOID(), SequenceNumber.next(), playerCount);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer is asking the player to stay or hit.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: Ask the player what he/she wants to do next.
     * Either get a another card (HIT) or not (STAY).</li>
     * <li>Clients: Dealer</li></ul>
     * @throws CommException
     */
    public int playerOption() throws CommException {

        int playerOption = 0;

        try {
            playerOption = remotePlayer.playerOption(
                    ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
        return playerOption;
    }

    /**
     * The dealer is telling the player that the game if over.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: Let the player know that the game is over, if he has won,
     *           and his new chip balance.</li>
     * <li>Clients: Dealer</li></ul>
     * @param won True iff the player has won.
     * @param newBalance The new chip balance of the player.
     * @throws CommException
     */
    public void gameOver(boolean won, int newBalance)
            throws CommException {

        try {
            remotePlayer.gameOver(ObjectID.getOID(), SequenceNumber.next(),
                    won, newBalance);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer is kicking the player off the game for misbehavior
     * <ul><li>Category: Realtime </li>
     * <li>Behavior: If a player does not place a bet or choose an option
     *           within the allotted amout of time, the Dealer kicks
     *           him out.</li>
     * <li>Clients: Dealer</li></ul>
     * @throws CommException
     */
    public void kickOut() throws CommException {

        try {
            Watchdog wd = new Watchdog(Thread.currentThread(), 5);
            remotePlayer.kickOut(ObjectID.getOID(), SequenceNumber.next());
            wd.disable();
        } catch (InterruptedException e) {
            throw new CommException(e);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }


    // Private methods

    /**
     * Create a new player proxy.
     */
    private FTPlayerProxy(Player player) {

        remotePlayer = player;
    }


    // Fields

    /** The single instance of this class. */
    private static FTPlayerProxy theInstance = null;

    /** The player stub. */
    private Player remotePlayer = null;
}
