// $Id: PlayerProxy.java,v 1.7 2003/03/15 19:21:18 lw Exp $
package team5.casino.comm;

import org.omg.CORBA.SystemException;
import team5.base.comm.*;

/**
 * This class acts as in intermediary between the dealer and the player
 * CORBA stubs. All calls from the PlayerOperations interface are forwarded
 * to the corresponding Player stub.
 * Each player should have access to exactly one floor, therefore this
 * class is implemented as a singleton.
 *
 * @author Team 5
 * @version $Revision: 1.7 $
 */
public class PlayerProxy {

    // Constructor

    /**
     * Create a new player proxy for a remote player.
     * @param player reference to the remote player object.
     */
    public PlayerProxy(Player player) {

        remotePlayer = player;
    }


    // Public methods

    // Player operations

    // TODO: Fix comments for player operations
    /**
     * A dealer request a player to bet on his/her cards.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: The player is prompted for his bet for the next game.</li>
     * <li>Clients: Dealer</li></ul>
     * @return The client's bet in chips.
     * @throws CommException if there is any error.
     */
    public int placeBet() throws CommException {

        int bet = 0;

        try {
            bet = remotePlayer.placeBet(
                    ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
        return bet;
    }

    /**
     * The player receives a new card from the dealer.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: This method passes a card to the player.</li>
     * <li>Clients: Dealer</li></ul>
     * @param cardID the ID of the new card.
     * @throws CommException if there is any error
     */
    public void acceptCard(byte cardID) throws CommException {

        try {
            remotePlayer.acceptCard(
                    ObjectID.getOID(), SequenceNumber.next(), cardID);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer is giving a player another player's card to view
     * if there are several players at a table.
     * <ul><li>Category: Baseline</li>
     * <li>Clients: Dealer</li></ul>
     * @param playerNumber The number of the other player at the
     *                table who gets this card.
     * @param cardID The ID of the new card.
     * @throws CommException on any error
     */
    public void acceptOthersCard(int playerNumber, byte cardID)
            throws CommException {

        try {
            remotePlayer.acceptOthersCard(ObjectID.getOID(),
                    SequenceNumber.next(), playerNumber, cardID);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer tells the player that the game is starting now.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: Start a new game at the current table.</li>
     * <li>Clients: Dealer</li></ul>
     * @param playerCount The number of players at the table.
     * @throws CommException on any error
     */
    public void startGame(byte playerCount) throws CommException {

        try {
            remotePlayer.startGame(
                    ObjectID.getOID(), SequenceNumber.next(), playerCount);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer is asking the player to stay or hit.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: Ask the player what he/she wants to do next.
     * Either get a another card (HIT) or not (STAY).</li>
     * <li>Clients: Dealer</li></ul>
     * @return the option chosen by the player
     * @throws CommException
     */
    public int playerOption() throws CommException {

        int playerOption = 0;

        try {
            playerOption = remotePlayer.playerOption(
                    ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
        return playerOption;
    }

    /**
     * The dealer tells the player that the game if over.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: Let the player know that the game is over, if he has won,
     *           and his new chip balance.</li>
     * <li>Clients: Dealer</li></ul>
     * @param won True iff the player has won.
     * @param newBalance The new chip balance of the player.
     * @throws CommException on any error.
     */
    public void gameOver(boolean won, int newBalance)
            throws CommException {

        try {
            remotePlayer.gameOver(ObjectID.getOID(), SequenceNumber.next(),
                    won, newBalance);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The dealer is kicking the player off the game for misbehavior
     * <ul><li>Category: Realtime </li>
     * <li>Behavior: If a player does not place a bet or choose an option
     *           within the allotted amout of time, the Dealer kicks
     *           him out.</li>
     * <li>Clients: Dealer</li></ul>
     * @throws CommException on any error.
     */
    public void kickOut() throws CommException {

        try {
            remotePlayer.kickOut(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }


    // Fields

    /** The player stub. */
    private Player remotePlayer = null;
}
