// $Id: BankProxy.java,v 1.10 2003/03/10 02:17:42 lw Exp $
package team5.player.comm;

import org.omg.CORBA.SystemException;
import team5.base.comm.*;

import java.util.Properties;

/**
 * This class acts as in intermediary between the player and the bank
 * CORBA stubs. All calls from the BankOperations interface are forwarded
 * to the corresponding Bank stub.
 * Each player should have access to exactly one bank, therefore this
 * class implements the singleton pattern.
 *
 * @author Team 5
 * @version $Revision: 1.10 $
 */
public class BankProxy implements IBankProxy {

    // Public methods

    /**
     * Get the bank proxy object. Create it if necessary.
     * @return the bank proxy
     */
    public synchronized static BankProxy getInstance(Properties orbprops)
        throws CommException, DuplicateValueException {

        if (theInstance == null) {
            theInstance = new BankProxy(orbprops);
        }
        return theInstance;
    }


    // Bank operations

    /**
     * The player is purchasing chips from the bank.
     * <ul><li>Category: Baseline.</li>
     * <li>Behavior: The player asks for quantity of chips.
     * <li>Clients:  Player</li></ul>
     * @param quantity The number of chips requested.
     * @throws CommException
     * @throws InvalidQuantity
     */
    public void buyChips(int quantity) throws CommException, InvalidQuantity {

        try {
            remoteBank.buyChips(ObjectID.getOID(), SequenceNumber.next(), quantity);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The player is selling chips back to the bank.
     * <ul><li>Category: Baseline.</li>
     * <li>Behavior: The player sells all or some of his chips back to the
     *           Bank.</li>
     * <li>Clients:  Player</li></ul>
     * @param quantity The number of chips requested.
     * @throws CommException
     * @throws InvalidQuantity
     */
    public void sellChips(int quantity) throws CommException, InvalidQuantity {

        try {
            remoteBank.sellChips(ObjectID.getOID(), SequenceNumber.next(), quantity);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The player is selling chips back to the bank.
     * <ul><li>Category: Baseline.</li>
     * <li>Behavior: The player sells all or some of his chips back to the
     *           Bank.</li>
     * <li>Clients:  Player</li></ul>
     * @throws CommException
     */
    public void sellAllChips() throws CommException {

        try {
            remoteBank.sellAllChips(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The player requests the number of chips he/she owns
     * <ul><li>Category: Baseline.</li>
     * <li>Behavior: The player requests the balance of chips from the bank.</li>
     * <li>Clients:  Player</li></ul>
     * @return current number of chips
     * @throws CommException
     */
    public int getChipBalance(int playerID)
            throws CommException {

        try {
            return remoteBank.getChipBalance(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }


    // Private methods

    /**
     * Create a new BankProxy.
     */
    private BankProxy(Properties orbprops) throws CommException,DuplicateValueException {

        remoteBank = FloorProxy.getInstance(orbprops).getBank();
    }


    // Fields

    /** The single instance of this class. */
    private static BankProxy theInstance = null;

    /** The bank stub. */
    private Bank remoteBank = null;

}





