// $Id: FTFloorProxy.java,v 1.10 2003/03/15 18:42:49 lw Exp $
package team5.player.comm;

import java.util.Properties;

import org.omg.CORBA.ORB;
import org.omg.CORBA.SystemException;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.NotFound;
import org.omg.CosNaming.NamingContextPackage.InvalidName;

import team5.base.comm.*;
import team5.base.UnhandledException;

/**
 * This class acts as in intermediary between the player and the floor
 * CORBA stubs. All calls from the FloorOperations interface are forwarded
 * to the corresponding Floor stub.
 * Each player should have access to exactly one floor, therefore this
 * class is implemented as a singleton.
 *
 * @author Team 5
 * @version $Revision: 1.10 $
 */
public class FTFloorProxy implements IFloorProxy {

    // Constants
    public static final int NORMAL = 0;
    public static final int FAILOVER = 1;


    // Public methods

    /**
     * Get the floor proxy object. Create it if necessary.
     *
     * @return the floor proxy
     */
    public synchronized static FTFloorProxy getInstance(Properties orbprops)
            throws CommException, DuplicateValueException {

        if (theInstance == null) {
            theInstance = new FTFloorProxy(orbprops);
        }
        Verbose.println("In: FTFloorProxy.getInstance");
        return theInstance;
    }


    // Floor operations

    // TODO: Fix comments for floor operations
    /**
     * Returns an object ID to the caller. This OID is guaranteed to be unique and
     * can be used to identify players or tables.
     * <ul><li>Category: Fault tolerance</li>
     * <li>Behavior: Player clients and tables use this method to obtain a unique ID.
     * <li>Clients: Player, Table</li></ul>
     * @return long The new object ID.
     */
    public int getOID() throws CommException {

        int oid = 0;

        try {
            oid = remoteFloor.getOID();
        } catch (SystemException e) {
            try {
                getNewFloor();
                oid = remoteFloor.getOID();
            } catch (SystemException e1) {
                throw new CommException(e1);
            }
        }
        return oid;
    }

    /**
     * The new or returning player registers with the system to start
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to register for
     *               a session at the casino server.</li>
     * <li>Clients: Player</li></ul>
     * @param playerName The client's Player interface.
     * @return long The player's ID number
     * @throws team5.base.comm.CommException
     */
    public int register(String playerName) throws CommException, InvalidPlayerName {

        int playerID = 0;

        try {
            playerID = remoteFloor.register(ObjectID.getOID(), SequenceNumber.next(), playerName);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            try {
                getNewFloor();
                playerID = remoteFloor.register(ObjectID.getOID(), SequenceNumber.next(), playerName);
            } catch (InvalidObjectID ex) {
                throw new CommException(ex);
            } catch (InvalidSequenceNumber ex) {
                throw new CommException(ex);
            } catch (SystemException ex) {
                throw new CommException(ex);
            }
        }
        return playerID;
    }

    /**
     * The player asks to leave the game.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to unregister for a session
     *           from the casino server.</li>
     * <li>Clients: Player</li></ul>
     * @throws CommException
     */
    public void unregister() throws CommException {

        try {
            remoteFloor.unregister(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            try {
                getNewFloor();
                remoteFloor.unregister(ObjectID.getOID(), SequenceNumber.next());
            } catch (InvalidObjectID ex) {
                throw new CommException(ex);
            } catch (InvalidSequenceNumber ex) {
                throw new CommException(ex);
            } catch (SystemException ex) {
                throw new CommException(ex);
            }
        }
    }

    /**
     * The player requests the Bank interface.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to get the interface
     *           for banking allowing him/her to buy and sell chips.</li>
     * <li>Clients: Player</li></ul>
     * @return Bank The Banking interface.
     * @throws CommException
     */
    public Bank getBank(int mode) throws CommException {

        Bank bank = null;

        if (mode == FAILOVER) {
            getNewFloor();
        }

        try {
            bank = remoteFloor.getBank(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            if (mode == NORMAL) {
                try {
                    getNewFloor();
                    bank = remoteFloor.getBank(ObjectID.getOID(), SequenceNumber.next());
                } catch (InvalidObjectID ex) {
                    throw new CommException(ex);
                } catch (InvalidSequenceNumber ex) {
                    throw new CommException(ex);
                } catch (CommException ex) {
                    throw new CommException(ex);
                }
            } else {
                throw new CommException(e);
            }
        }

        return bank;
    }

    /**
     * The player requests the Tables interface.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to get the interface
     *           to allow him/her to join a game</li>
     * <li>Clients: Player</li></ul>
     * @return Bank The Tables interface.
     * @throws CommException
     */
    public Tables getTables(int mode) throws CommException {

        Tables tables = null;

        if (mode == FAILOVER) {
            getNewFloor();
        }

        try {
            tables = remoteFloor.getTables(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            if (mode == NORMAL) {
                try {
                    getNewFloor();
                    tables = remoteFloor.getTables(ObjectID.getOID(), SequenceNumber.next());
                } catch (InvalidObjectID ex) {
                    throw new CommException(ex);
                } catch (InvalidSequenceNumber ex) {
                    throw new CommException(ex);
                } catch (SystemException ex) {
                    throw new CommException(ex);
                }
            } else {
                throw new CommException(e);
            }
        }
        return tables;
    }


    // Private methods

    /**
     * Create a new floor proxy. This also initializes the orb and assigns
     * an object id to the client.
     * @throws CommException
     * @throws team5.base.comm.DuplicateValueException
     */
    private FTFloorProxy(Properties orbprops)
            throws CommException, DuplicateValueException {

        Verbose.println("In FTFloorProxy.<init>");
        // Initialize the orb
        orb = ORB.init(new String[0], orbprops);
        try {
            // Contact the Naming Service
            org.omg.CORBA.Object objRef =
                    orb.resolve_initial_references("NameService");
            nsRef = NamingContextExtHelper.narrow(objRef);
        } catch (SystemException e) {
            throw new CommException(e);
        } catch (org.omg.CORBA.ORBPackage.InvalidName e) {
            throw new CommException(e);
        }
        this.orbprops = orbprops;

        // Lookup the floor in the Naming Service
        try {
            remoteFloor = FloorHelper.narrow(nsRef.resolve_str("Floor"));
        } catch (NotFound e) {
            throw new CommException("No Floor available");
        } catch (CannotProceed e) {
            throw new CommException("Could not connect to floor", e);
        } catch (InvalidName e) {
            throw new UnhandledException(e);
        }

        // Get object id and store for future reference
        ObjectID.setOID(remoteFloor.getOID());
    }

    /**
     * Get a new remote floor
     */
    private void getNewFloor() throws CommException {

        boolean foundFloor = false;

        for (int i = 0; !foundFloor && i < 10; i++) {
            // Lookup the floor in the Naming Service
            try {
                remoteFloor = FloorHelper.narrow(nsRef.resolve_str("Floor"));
                foundFloor = true;
            } catch (NotFound e) {
            } catch (CannotProceed e) {
            } catch (InvalidName e) {
            }

            try {
                if (!foundFloor) {
                    Thread.sleep(1000);
                }
            } catch (InterruptedException e) {
            }
        }

        // Throw exception if we still don't have a floor
        if (!foundFloor) {
            throw new CommException("No Floor available");
        }
        // Set new remote bank and tables
        try {
            FTBankProxy.getInstance(orbprops).remoteBank = remoteFloor.
                    getBank(ObjectID.getOID(), SequenceNumber.next());
            FTTablesProxy.getInstance(orbprops).remoteTables = remoteFloor.
                    getTables(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        }
    }

    // Fields

    /** The single instance of this class. */
    private static FTFloorProxy theInstance = null;

    /** The floor stub. */
    private Floor remoteFloor = null;

    /** The orb. */
    private ORB orb = null;

    /** The naming context */
    private NamingContextExt nsRef = null;

    private Properties orbprops;
}
