// $Id: FloorProxy.java,v 1.6 2003/03/10 02:04:01 lw Exp $
package team5.player.comm;

import java.util.Properties;

import org.omg.CORBA.ORB;
import org.omg.CORBA.SystemException;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.NotFound;
import org.omg.CosNaming.NamingContextPackage.InvalidName;

import team5.base.comm.*;

/**
 * This class acts as in intermediary between the player and the floor
 * CORBA stubs. All calls from the FloorOperations interface are forwarded
 * to the corresponding Floor stub.
 * Each player should have access to exactly one floor, therefore this
 * class is implemented as a singleton.
 *
 * @author Team 5
 * @version $Revision: 1.6 $
 */
public class FloorProxy implements IFloorProxy {

    // Public methods

    /**
     * Get the floor proxy object. Create it if necessary.
     *
     * @return the floor proxy
     */
    public synchronized static FloorProxy getInstance(Properties orbprops)
            throws CommException, DuplicateValueException {

        if (theInstance == null) {
            theInstance = new FloorProxy(orbprops);
        }
        return theInstance;
    }


    // Floor operations

    // TODO: Fix comments for floor operations
    /**
     * Returns an object ID to the caller. This OID is guaranteed to be unique and
     * can be used to identify players or tables.
     * <ul><li>Category: Fault tolerance</li>
     * <li>Behavior: Player clients and tables use this method to obtain a unique ID.
     * <li>Clients: Player, Table</li></ul>
     * @return long The new object ID.
     */
    public int getOID() throws CommException {

        try {
            return remoteFloor.getOID();
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The player requests the Bank interface.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to get the interface
     *           for banking allowing him/her to buy and sell chips.</li>
     * <li>Clients: Player</li></ul>
     * @return Bank The Banking interface.
     * @throws CommException
     */
    public Bank getBank() throws CommException {

        try {
            return remoteFloor.getBank(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The player requests the Tables interface.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to get the interface
     *           to allow him/her to join a game</li>
     * <li>Clients: Player</li></ul>
     * @return Bank The Tables interface.
     * @throws CommException
     */
    public Tables getTables() throws CommException {

        try {
            return remoteFloor.getTables(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The new or returning player registers with the system to start
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to register for
     *               a session at the casino server.</li>
     * <li>Clients: Player</li></ul>
     * @param playerName The client's Player interface.
     * @return long The player's ID number
     * @throws team5.base.comm.CommException
     */
    public int register(String playerName) throws CommException, InvalidPlayerName {

        try {
            return remoteFloor.register(ObjectID.getOID(), SequenceNumber.next(), playerName);
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }

    /**
     * The player asks to leave the game.
     * <ul><li>Category: Baseline</li>
     * <li>Behavior: A player client uses this method to unregister for a session
     *           from the casino server.</li>
     * <li>Clients: Player</li></ul>
     * @throws CommException
     */
    public void unregister() throws CommException {

        try {
            remoteFloor.unregister(ObjectID.getOID(), SequenceNumber.next());
        } catch (InvalidObjectID e) {
            throw new CommException(e);
        } catch (InvalidSequenceNumber e) {
            throw new CommException(e);
        } catch (SystemException e) {
            throw new CommException(e);
        }
    }


    // Private methods

    /**
     * Create a new floor proxy. This also initializes the orb and assigns
     * an object id to the client.
     * @throws CommException
     * @throws team5.base.comm.DuplicateValueException
     */
    private FloorProxy(Properties orbprops) throws CommException, DuplicateValueException {

        NamingContextExt nsRef = null;
        // Initialize the orb
        orb = ORB.init(new String[0], orbprops);
        try {
            // Contact the Naming Service
            org.omg.CORBA.Object objRef =
                    orb.resolve_initial_references("NameService");
            nsRef = NamingContextExtHelper.narrow(objRef);
        } catch (SystemException e) {
            throw new CommException(e);
        } catch (org.omg.CORBA.ORBPackage.InvalidName e) {
            throw new CommException(e);
        }

        // Lookup the floor in the Naming Service
        try {
            remoteFloor = FloorHelper.narrow(nsRef.resolve_str("Floor"));
        } catch (NotFound e) {
            throw new CommException(e);
        } catch (CannotProceed e) {
            throw new CommException(e);
        } catch (InvalidName e) {
            throw new CommException(e);
        }

        // Get object id and store for future reference
        ObjectID.setOID(remoteFloor.getOID());
    }


    // Fields

    /** The single instance of this class. */
    private static FloorProxy theInstance = null;

    /** The floor stub. */
    private Floor remoteFloor = null;

    /** The orb. */
    private ORB orb = null;
}

