// $Id: Repman.java,v 1.10 2003/03/15 00:48:59 lw Exp $
package team5.repman;

import gnu.getopt.Getopt;
import org.omg.CORBA.ORB;
import org.omg.CORBA.SystemException;
import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.InvalidName;
import org.omg.CosNaming.NamingContextPackage.NotFound;
import team5.base.comm.CommException;
import team5.base.comm.Floor;
import team5.base.comm.FloorHelper;
import team5.base.comm.Verbose;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.*;

/**
 * This class must be run to start up the casino server.
 *
 * @author Team 5
 * @version $Revision: 1.10 $
 */
public class Repman {

    // Public methods

    /**
     * Main method for Repman class
     * <ul><li>Category: Fault-Tolerance </li>
     * <li>Behavior: Creates all the server replicase and maintains
     * the Corba namespace</li>
     * @param args The startCommand line arguments.
     */
    public static void main(String[] args) {

        // Setup Shutdown thread to catch kill signals
        Runtime rt = Runtime.getRuntime();
        Repman repman = new Repman();
        ShutdownThread shutdownThread = new ShutdownThread(repman);
        rt.addShutdownHook(shutdownThread);

        repman.startup(args);
    }

    public Repman() {
    }

    public synchronized void startup(String[] args) {

        Properties props;
        Iterator i;
        String host = "";
        int count;
        ORB orb;

        //set defaults
        sleeptime = 1000;//one second;

        parseArgs(args);
        props = loadProps();
        banner();
        // ORB initialization
        orb = ORB.init(args, props);
        Verbose.println("---Created ORB instance.");
        if (!findNameService(orb)) {
            System.err.println("Info: Start the name service before trying again");
            return;
        }

        Verbose.println("---Connected to naming service.");

        if (availableHosts.size() == 0) {
            System.err.println("Fatal: No hostnames available - Forget it");
            System.exit(1);
        }

        // try to start numReplica replicas on hosts from the list
        count = 0;
        i = ((List) availableHosts.clone()).iterator();
        do {
            if (i.hasNext()) {
                host = (String) i.next();
            } else {
                System.err.println("Warning: Not enough hosts available to start servers on");
                System.err.println("         Started " + count + " replicas");
                numReplicas = count;
                break;
            }
            try {
                Replica replica = new Replica(this, namingContext,
                        host, count, sleeptime, startCommand);
                // replica successfully started
                availableHosts.remove(host);
                replicas.put(host, replica);
                replica.start();
                count++;
            } catch (InterruptedException e) {
                System.err.println("Warning: Could not start server on " + host);
                System.err.println("         Deleting " + host + " from list of hosts");
                availableHosts.remove(host);
            } catch (CommException e) {
                System.err.println("Fatal: Error contacting name service - exiting. " + e);
                cleanup();
                System.exit(1);
            }
        } while (count < numReplicas && !shutdown);

        if (!shutdown) {
            if (host != "") {
                setPrimary(host);
            } else {
                System.err.println("Fatal: Internal error: hostname empty");
                cleanup();
                System.exit(-1);
            }
        }
    }

    public synchronized void replacePrimary(Replica crashed) {

        String crashedHost = crashed.getHostName();
        String newHost = "";
        Iterator i;

        // find a host with a running replica
        i = replicas.keySet().iterator();
        while (i.hasNext()) {
            String h = (String) i.next();

            if (h != crashedHost) {
                newHost = h;
            }
        }
        if (newHost != "") {
            setPrimary(newHost);
        } else {
            System.err.println("Fatal: No other running replica available");
            cleanup();
            System.exit(-1);
        }
    }

    public synchronized void setPrimary(String host) {

        Replica newPrimary = (Replica) replicas.get(host);
        String primaryName = "Floor" + newPrimary.getNumber();
        NameComponent[] nc = {new NameComponent("Floor", "")};
        Floor primary = null;

        try {
            primary = FloorHelper.narrow(namingContext.resolve_str(primaryName));
            namingContext.rebind(nc, primary);
            System.out.println("Info: " + primaryName + " is the new primary");
        } catch (NotFound e) {
            System.err.println("Fatal: Could not set new primary.");
            System.err.println("       Name not found: " + primaryName);
            cleanup();
            System.exit(-1);
        } catch (CannotProceed e) {
            System.err.println("Fatal: Could not set new primary.");
            System.err.println("       Problems contacting nameservice");
            cleanup();
            System.exit(-1);
        } catch (InvalidName e) {
            System.err.println("Fatal: Could not set new primary.");
            System.err.println("       Invalid name: " + primaryName);
            cleanup();
            System.exit(-1);
        } catch (SystemException e) {
            System.err.println("Fatal: Could not set new primary.");
            System.err.println("       CORBA failure" + e);
            cleanup();
            System.exit(-1);
        }
        try {
            newPrimary.activate();
        } catch (CommException e) {
            System.err.println("Fatal: Could not activate new primary.");
            System.err.println("       CORBA failure" + e);
            cleanup();
            System.exit(-1);
	}
    }

    /**
     * This method is called by a replica object if the pinged process
     * doesn't respond any more.
     */
    public synchronized void notifyReplicaDeath(Replica deadReplica) {

        String host = deadReplica.getHostName();
        int number = deadReplica.getNumber();

        Runtime rt = Runtime.getRuntime();
        Iterator i = replicas.keySet().iterator();
        String osName = System.getProperty("os.name");
        NameComponent[] nc = new NameComponent[1];
        String name = "Floor" + number;
        String[] cmd = new String[5];

        // clean up nameing service
        try {
            Verbose.println("---Unbinding '" + name + "'");
            nc[0] = new NameComponent(name, "");
            namingContext.unbind(nc);
        } catch (NotFound e) {
            System.err.println("Error: Could not unbind name '" + name + "'");
        } catch (CannotProceed e) {
            System.err.println("Error: Could not unbind name '" + name + "'");
        } catch (InvalidName e) {
            System.err.println("Error: Could not unbind name '" + name + "'");
        } catch (SystemException e) {
            System.err.println("Error: Could not unbind name '" + name + "'");
        }
        // kill the process to be sure it's gone
        if (osName.equals("Windows 2000") || osName.equals("Windows XP")) {
            cmd[0] = "cmd.exe";
            cmd[1] = "/C";
            cmd[2] = "C:\\cygwin\\bin\\ssh ";
            cmd[3] = "lw@" + host + ".ece.cmu.edu ";
            cmd[4] = killCommand;
        } else if (osName.equals("Linux")) {
            cmd[0] = "ssh";
            cmd[1] = host;
            cmd[2] = killCommand;
            cmd[3] = "";
            cmd[4] = "";
        } else {
            Verbose.println("Fatal: Unknown Operating System " + osName);
            cleanup();
            System.exit(-1);
        }

        Verbose.println("---Killing replica on " + host);
        Verbose.println("---Execing " + cmd[0] + " " + cmd[1] + " " + cmd[2] +
                cmd[3] + cmd[4]);
        try {
            Process p = rt.exec(cmd);
            p.waitFor();
        } catch (Exception e) {
            // ignore!
            System.err.println("Error: Could not kill replica on " + host);
        }

        // try to create new replica
        // first on same host
        boolean success = false;
        try {
            Replica replica = new Replica(this, namingContext,
                    host, number, sleeptime, startCommand);
            // replica successfully started
            availableHosts.remove(host);
            replicas.put(host, replica);
            replica.start();
            success = true;
        } catch (InterruptedException e) {
            System.err.println("Warning: Could not start server on " + host);
            System.err.println("         Deleting " + host + " from list of hosts");
            availableHosts.remove(host);
        } catch (CommException e) {
            System.err.println("Fatal: Error contacting name service - exiting." + e);
            cleanup();
            System.exit(1);
        }
        Thread.yield();

        // now try other host
        Iterator h = ((List) availableHosts.clone()).iterator();
        while (!success && h.hasNext() && !shutdown) {
            host = (String) i.next();
            try {
                Replica replica = new Replica(this, namingContext,
                        host, number, sleeptime, startCommand);
                // replica successfully started
                availableHosts.remove(host);
                replicas.put(host, replica);
                replica.start();
                success = true;
            } catch (InterruptedException e) {
                System.err.println("Warning: Could not start server on " + host);
                System.err.println("         Deleting " + host + " from list of hosts");
                availableHosts.remove(host);
            } catch (CommException e) {
                System.err.println("Fatal: Error contacting name service - exiting." + e);
                cleanup();
                System.exit(1);
            }
            Thread.yield();
        }

        if (!shutdown) {
            if (!success) {
                System.err.println("Error: No host available to restart replica");
                numReplicas--;
            }
            Verbose.println("---New replica started on " + host);
        }
    }

    /**
     * Kill all replica and cleanup name service.
     */
    public synchronized void cleanup() {

        Runtime rt = Runtime.getRuntime();
        Iterator i = replicas.keySet().iterator();
        NameComponent[] nc = new NameComponent[1];
        String osName = System.getProperty("os.name");
        String[] cmd = new String[5];
        String host = "";

        shutdown = true;

        // are we shutting down due to missing name service?
        if (namingContext == null) {
            return;
        }

        // clean up the naming service
        for (int k = -1; k < numReplicas; k++) {
            String name = "Floor";

            if (k != -1) {
                name += k;
            }
            try {
                Verbose.println("---Unbinding '" + name + "'");
                nc[0] = new NameComponent(name, "");
                namingContext.unbind(nc);
            } catch (NotFound e) {
                System.err.println("Error: Could not unbind name '" + name + "'");
            } catch (CannotProceed e) {
                System.err.println("Error: Could not unbind name '" + name + "'");
            } catch (InvalidName e) {
                System.err.println("Error: Could not unbind name '" + name + "'");
            } catch (SystemException e) {
                System.err.println("Error: Could not unbind name '" + name + "'");
            }
        }

        if (osName.equals("Windows 2000") || osName.equals("Windows XP")) {
            cmd[0] = "cmd.exe ";
            cmd[1] = "/C ";
            cmd[2] = "C:\\cygwin\\bin\\ssh ";
            cmd[3] = "lw@" + host + ".ece.cmu.edu \"";
            cmd[4] = killCommand;
        } else if (osName.equals("Linux")) {
            cmd[0] = "ssh";
            cmd[1] = host;
            cmd[2] = killCommand;
            cmd[3] = "";
            cmd[4] = "";
        } else {
            Verbose.println("Fatal: Unknown Operating System " + osName);
            System.exit(-1);
        }

        // kill all replicas
        while (i.hasNext()) {
            // set hostname
            host = (String) i.next();
            if (osName.equals("Windows 2000") || osName.equals("Windows XP")) {
                cmd[3] = "lw@" + host + ".ece.cmu.edu ";
            } else if (osName.equals("Linux")) {
                cmd[1] = host;
            }
            Verbose.println("---Killing replicas on " + host);
            Verbose.println("---Execing " + cmd[0] + " " + cmd[1] + " " + cmd[2] +
                    cmd[3] + cmd[4]);
            try {
                Process p = rt.exec(cmd);
                p.waitFor();
            } catch (Exception e) {
                // ignore!
                System.err.println("Error: Could not kill replica on " + host);
            }
        }
    }


    // Private methods

    /**
     * Start a new replica on an available host.
     *
     * @param number the number of the new replica
     * @param host the name of the host to start the server on. If hostname is empty
     *             or null select a host from the list of available hosts.
     * @return whether the operation was successful
     */
    /*
    private boolean startAnotherReplica(String host) {

        boolean success = false;

        if (hostname != null && !hostname.equals("")) {
            // first on passed host
            try {
                Replica replica = new Replica(this, namingContext,
                        host, number, sleeptime, startCommand);
                // replica successfully started
                availableHosts.remove(host);
                replicas.put(host, replica);
                replica.start();
                success = true;
            } catch (InterruptedException e) {
                System.err.println("Warning: Could not start server on " + host);
                System.err.println("         Deleting " + host + " from list of hosts");
                availableHosts.remove(host);
            } catch (CommException e) {
                System.err.println("Fatal: Error contacting name service - exiting." + e);
                cleanup();
                System.exit(1);
            }
            Thread.yield();
        }

        // now try other host
        Iterator h = ((List) availableHosts.clone()).iterator();
        while (!success && h.hasNext() && !shutdown) {
            host = (String) i.next();
            try {
                Replica replica = new Replica(this, namingContext,
                        host, number, sleeptime, startCommand);
                // replica successfully started
                availableHosts.remove(host);
                replicas.put(host, replica);
                replica.start();
                success = true;
            } catch (InterruptedException e) {
                System.err.println("Warning: Could not start server on " + host);
                System.err.println("         Deleting " + host + " from list of hosts");
                availableHosts.remove(host);
            } catch (CommException e) {
                System.err.println("Fatal: Error contacting name service - exiting." + e);
                cleanup();
                System.exit(1);
            }
            Thread.yield();
        }
        return success;
    }
    */

    /**
     * Obtain a reference to the nameing context.
     *
     * @param orb the orb instance to contact
     */
    private boolean findNameService(ORB orb) {

        boolean success = false;

        try {
            namingContext = NamingContextExtHelper.narrow(
                    orb.resolve_initial_references("NameService"));
            success = true;
        } catch (org.omg.CORBA.ORBPackage.InvalidName e) {
            System.err.println("Fatal: Naming service not found");
        } catch (SystemException e) {
            System.err.println("Fatal: Could not contact naming service");
        }
        return success;
    }

    /**
     * Load the properties file from disk. Properties are
     * <p/>
     * <table>
     *   <th>
     *     <td>Property name</td><td>Description</td>
     *   </th>
     *   <tr>
     *     <td>repman.host.list</td><td>A space separated list of hostnames that can run replicas</td>
     *   </tr>
     *   <tr>
     *     <td>repman.replica.count</td><td>The desired number of replicas</td>
     *   </tr>
     *   <tr>
     *     <td>repman.start.command</td><td>Command line passed to ssh to start a replica</td>
     *   </tr>
     *   <tr>
     *     <td>repman.kill.command</td><td>Command line passed to ssh to kill a replica</td>
     *   </tr>
     * </table>
     * @return the properties object.
     */
    private Properties loadProps() {

        Properties props = null;
        String propName = null;

        try {
            props = new Properties(System.getProperties());
            props.load(new BufferedInputStream(new FileInputStream(propfileName)));
        } catch (FileNotFoundException e) {
            System.out.println("Fatal error: Could not find property file " + propfileName);
            System.exit(-1);
        } catch (IOException e) {
            System.out.println("Fatal error: Could not read property file " + propfileName);
            System.exit(-1);
        }

        // enter hostnames
        propName = "repman.host.list";

        String hostListString = props.getProperty(propName);
        String[] hostList = hostListString.split(" ");

        if (hostListString == null) {
            System.err.println("Fatal error: Could not read property " + propName);
            System.exit(-1);
        }
        for (int i = 0; i < hostList.length; i++) {
            availableHosts.add(hostList[i]);
        }
        Verbose.println("---" + propName + "=" + hostListString);

        // get desired number of replicas
        propName = "repman.replica.count";

        try {
            numReplicas = (new Integer(props.getProperty(propName))).intValue();
        } catch (NumberFormatException e) {
            System.err.println("Fatal error: Could not read property" + propName);
            System.exit(-1);
        }
        if (numReplicas > hostList.length) {
            numReplicas = hostList.length;
            System.err.println("Warning: Only " + numReplicas + "host names specified");
            System.err.println("         Changing number of replicas to " + numReplicas);
        }
        if (numReplicas == 0) {
            System.err.println("Fatal error: 0 replicas requested - Nice try.");
            System.exit(-1);
        }
        Verbose.println("---" + propName + "=" + numReplicas);

        // get start command string
        propName = "repman.start.command";

        startCommand = props.getProperty(propName);
        if (startCommand == null) {
            System.err.println("Fatal error: Could not read property" + propName);
            System.exit(-1);
        }
        Verbose.println("---" + propName + "=" + startCommand);

        // get kill command string
        propName = "repman.kill.command";

        killCommand = props.getProperty(propName);
        if (killCommand == null) {
            System.err.println("Fatal error: Could not read property" + propName);
            System.exit(-1);
        }
        Verbose.println("---" + propName + "=" + killCommand);

        return props;
    }

    private void parseArgs(String[] argv) {

        Verbose.verbose = false;
        int c;
        Getopt g = new Getopt("Server", argv, "hp:s:n:v");

        while ((c = g.getopt()) != -1) {
            switch (c) {
                case 'h':
                    usage();
                    break;
                case 'p':
                    propfileName = g.getOptarg();
                    break;
                case 's':
                    sleeptime = new Integer(g.getOptarg()).intValue();
                    break;
                case 'n':
                    numReplicas = new Integer(g.getOptarg()).intValue();
                    break;
                case 'v':
                    System.out.println("---Setting Verbose Mode");
                    Verbose.verbose = true;
                    break;
                default:
                    System.out.println("Unknown Option " + g.getOptarg());
                    usage();
            }
        }
    }

    /**
     * Printout the commandline options.
     */
    private void usage() {

        System.out.println("Usage: java -classpath $CLASSPATH team5.casino.Server");
        System.out.println("-h Print this message.");
        System.out.println("-p <prop file name> Set the properties file");
        System.out.println("-v Set the output to verbose");
        System.out.println("-s Set the ping interval");
        System.exit(0);
    }

    /**
     * Prints a banner message in verbose mode.
     */
    private void banner() {
        Verbose.println("\n**************************************");
        Verbose.println(  "* Virtual Casino Replication Manager *");
        Verbose.println(  "**************************************\n");
    }


    // Fields

    /** A list of available host names to run replicas on. */
    private ArrayList availableHosts = new ArrayList();

    /**
     * This map keeps track of the running replicas. The key of each entry
     * is a host name and the value is the Replica object for this host.
     */
    private HashMap replicas = new HashMap();

    /** The command line string that must be executed to start a replica on a host. */
    private String startCommand = "";

    /** The command line string that must be executed to kill all replica on a host. */
    private String killCommand = "";

    /** Reference to the naming service. */
    private NamingContextExt namingContext;

    /** The name of the property file. */
    private String propfileName;

    /** The time between two pings to a replica. */
    private int sleeptime;

    /** The number of replicas that should be running at any time. */
    private int numReplicas = 0;

    /** Set to indicate shutdown. */
    private boolean shutdown = false;
}
